
#####################################################################################
################################## IMPORTS PYTHON3 ##################################
#####################################################################################

from module_formule import *
import os.path #gére la création du fichier de sortie du simulateur
import math #permet de gérer l'infini
import sys #permet de gérer les manipulations d'arguments en entrée du programme.

##############################################################################################
################################## DÉFINITION DES FONCTIONS ##################################
##############################################################################################

def pred(v,G): #prend en entrée un graphe d'interaction G et retourne en sortie la liste des prédécesseurs de la variable v en entrée. 
	pred=[]
	for s in G: #pour chaque sommet du graphe,
		for e in G[s]: #on parcourt chacun de ses fils,
			if e==v: #si la variable en entrée est un fils de s, 
				pred=pred+[s] #alors s est le prédecesseur de v. 
	return pred	

def calcul_ressources(v,M,G,eta): #étant donné un graphe G et un ensemble de multiplexes M (en entrée), la fonction retourne (en sortie) l'ensemble des ressources d'une variable v (en entrée) pour un état donnée (en entrée).
	r=[]
	ress=pred(v,G) #appel de la fonction pred récupérant l'ensemble des ressources de la variable v.
	for m in M: #pour chaque multiplexe,
		if m in ress: #si le multiplexe m fait partie des ressources de v alors, 
			Fm=M[m] #récupération de la formule sous forme de chaîne de caractère correspondant au multiplexe m,
			lm=liste_lexicale(Fm) #appel de la fonction liste_lexicale dans le but de transformer la formule Fm en liste lexicale dans un soucis de malléabilité.
			am=mkfor(lm) #appel de la fonction mkfor. Elle construit un arbre syntaxique d'une formule sous forme de liste lexicale.
			res1=eval(am,eta) #appel de la fonction eval définissant la valeur de vérité (True ou False) d'une formule dans un état donné. 
			if res1==True: #si la formule est vraie dans l'état eta alors,
				r=r+[m] #on ajoute le multiplexe m dans la liste des ressources de la variable v. 
	if r==[]: #FLAG : on peut s'en passer
		r=r+[] #FLAG : on peut s'en passer
	r=tuple(r) #transformation de la liste des ressources de v en tuple. Le tuple a l'avantage d'être immuable.
	return r

def liste_variable(G): #fonction permettant de récupérer la liste des variables lvar d'un graphe G.
	lvar=[]
	for s in G: #pour chaque sommet du graphe,
		for f in G[s]: #parcourt de la liste des fils du sommet s, s étant un nom de multiplexe et G[s] la liste des cibles du multiplexe s. 
			if f not in lvar: #si f ne fait pas déjà partie de la liste de variable lvar alors,
				lvar=lvar+[f] #on ajoute f dans la liste des variables du graphe.
	return lvar

def discrete_levels(D,v): #fonction générant en sortie la liste n des niveaux discrets d'une variable v en entrée en considérant son niveau maximal définit dans le dictionnaire D en entrée. 
	n=[]
	dl=D[v] #d1 le niveau discret maximal de la variable v. 
	while dl>0: #pour être en mesure de considérer tous les niveaux discrets possibles, l'idée est de partir du niveau maximal et de remplir la liste n par décrémentation de dl. La boucle while s'interrompt lorsque dl=0. 
		n=n+[dl]
		dl=dl-1
	n=n+[0] #ajout du niveau 0 en fin de liste.
	return n

def celerite_par_etat(G,eta,M,D): #fonction retournant en sortie un dictionnaire. L'ensemble des variables v forment l'ensemble des clés du dictionnaire cele. Chaque variable prend pour valeur sa celérité correspondante dans un état donné.
	cele={}
	var=liste_variable(G) #var retourne la liste des variables du graphe G,
	for v in var:
		ress=calcul_ressources(v,M,G,eta) #ress retourne la liste des ressources pour la variable v considérée,
		nx=discrete_levels(D,v) #nx récupère le tuple des niveaux discrets de la variable v,
		for n in nx: #parcourt de l'ensemble des niveaux discrets de v,
			if n==eta[v]: #cette condition permet de récupérer uniquement les célérités indexées par un niveau discret correspondant à l'état courant considéré,
				lc=(v,ress,n) #construction de la célérité sous forme de triplet avec v le nom de la variable sous forme de chaîne de caractères, ress l'ensemble des ressources de v sous forme de tuple et n le niveau discret, de type entier, de la variable v.
		cele[v]=lc 
	return cele

def nx_parvar(G,D): #la fonction nx_parvar retourne la liste des niveaux discrets (valeur) de chaque variable (clé) sous la forme d'un dictionnaire dnx. La fonction nx_parvar est appelée uniquement dans la fonction extraction.
	dnx={}
	for v in D:
		nx=discrete_levels(D,v) #nx la liste des niveaux discrets d'une variable v,
		dnx[v]=nx #à chaque variable est associé la liste des niveaux discrets correspondants.
	return dnx 

def extraction(G,D): #fonction dont le but est de construire toutes les combinaisons possibles de liste(s) de n niveau(x) discret(s) de n variable(s). On considère un graphe d'état défini comme un espace à n dimension(s) (n étant le nombre de variable/sommet du graphe d'interaction) regroupant tous les états possibles du système étudié.
	dnx=nx_parvar(G,D) #dnx un dictionnaire avec pour clé une variable v et pour valeur correspondante la liste des niveaux discrets de v,
	ldico=[]
	for key,value in dnx.items(): #parcourt de chaque couple key : value du dictionnaire dnx (variable : liste des niveaux discrets), 
		ldico=prod_cart(ldico,key,value) #appel de la fonction prod_cart. Cette fonction calcule le produit cartésien de la liste des niveaux discrets par elle même (carré cartésien).
	return ldico #liste de dictionnaires. Chaque élément de la liste est un état sous forme de dictionnaire appartenant au graphe d'état.

def prod_cart(ldico,key,value): #la fonction prod-cart calcule le carré cartésien de la liste ldico (carré cartésien : produit cartésien d'un ensemble par lui même).
	dico={}
	l=[] #initialisation de L si la fonction prod_cart est appelée pour la première fois. Sinon, il s'agit d'une réinitialisation de la liste l. 
	if len(ldico)==0: #initialisation de la liste l si on appelle pour la 1ère fois la fonction prod_cart (liste de dictionnaire retournée en sortie). Si la liste ldico est vide, 
		for e in value: #alors on parcourt les valeurs de la clé (variable v),
			dico[key]=e #on associe à la variable v le niveau discret e,
			l.append(dico) #on stocke le couple variable v : niveau discret e dans la liste l. Exemple d'initialisation de la liste l par la variable v bivaluée : l=[{'v':0},{'v':1}]. La variable permettant l'initialisation de l est systématiquemnt le premier élément (key) du dictionnaire dnx (voir fonction extraction).
			dico={} #Une fois le contenu du dictionnaire dico stocké dans l, réinitialisation de dico.
	else: #si la liste ldico est non vide, alors calcul du carré cartésien de la liste de dictionnaire ldico dans le bloc suivant : 
		for f in ldico: #pour chaque dictionnaire de la liste ldico,
			for e in value: #pour chaque niveau discret de la variable key, 
				f1=f.copy() #copie préalable du dictionnaire f afin de manipuler un dictionnaire non altéré par l'ajout d'un élément,
				f1[key]=e #ajout du couple variable key et niveau discret de key dans le dictionnaire f,
				l=l+[f1] #ajout de f dans l.
	return l #retour de la liste de dictionnaire l stockée dans la variable ldico dans la fonction extraction. La longueur de la liste ldico est multipliée par le nombre de niveau discret de la variable key. De plus, la longueur de chaque dictionnaire de ldico augmente de 1 à chaque appel de la fonction prod_cart dans le parcourt du couple key,value (ligne 208) - (explicaion : ajout d'une variable et d'une valeur (ex : 'v':0)). 


def celerites_etats(G,M,D): #la fonction celerites_etats retourne la liste des célérités dans un état donné. 
	letat=extraction(G,D) #appel de la fonction extraction. Stockage de la liste de tous les états d'un graphe d'état à n dimensions (n le nombre de variable), sous forme de liste de dictionnaire, dans letat.
	listece=[] 
	for e in letat: #pour chaque état sous forme de dictionnaire dans la liste des états letat, 
		lce=celerite_par_etat(G,e,M,D) #appel de la fonction celerite_par_etat et stockage du dictionnaire en sortie dans lce (clé : variable du graphe d'interaction et valeur : liste des célérités de la variable dans un état donné),
		for s in lce: #pour chaque variable (entité biologique du graphe d'interaction), 
			listece=listece+[lce[s]] #récupération de la liste des célérités de la variable dans un état donné, et stockage des éléments de la liste dans listece,
	return listece #renvoi de la liste des célérités dans un état donné. 

def liste_cele(G,M,D): #la fonction liste_cele construit la liste des célérités pour un graphe d'interaction donné. Elle fait intervenir l'utilisateur au moment de la construction du dictionnaire dicoc. L'utilisateur fournit la valeur de chaque célérité et la fonction retourne dicoc sous la forme {'célérité': valeur réelle}.
	lcelerite=[] #initialisation de la liste des célérités 
	liste_etat_cel=celerites_etats(G,M,D) #stockage de la liste des célérités dans un état donné dans liste_etat_cel,
	for e in liste_etat_cel: #pour chaque célérité,
		if e not in lcelerite: #si la célérité e ne se trouve pas déjà dans lcelerite, 
			lcelerite=lcelerite+[e] #alors on l'ajoute dans la liste des célérités,
	print(len(lcelerite)," célérités.")
	dicoc={} #initialisation du dictionnaire dicoc stockant pour chaque célérité, sa valeur réélle (flottant) corespondante.
	for e in lcelerite: #pour chaque célérité,
		print("Célérité : ",e)
		val=input("Veuillez entrer la valeur de la célérité : ") #intervention de l'utilisateur pour l'entrée de la valeur de la célérité considérée, 
		val=float(val) #la fonction input renvoie toujours une chaîne de caractère donc conversion du type chaîne de caractère en type flottant. 
		dicoc[e]=val #remplissage du dictionnaire dicoc: à chaque célérité (e) est associée sa valeur correspondante (val).
	return dicoc #dicoc un dictionnaire assorciant à une célérité, sa valeur réelle (flottant).

def sgn(Ce): #la fonction sgn abstrait la valeur de la célérité à -1 si elle est strictement négative, à +1 si elle est strictement positive et à 0 si elle est nulle.
	if Ce<0: 
		sgnCe=-1 
	elif Ce>0:
		sgnCe=1
	elif Ce==0:
		sgnCe=0
	return sgnCe

def calcul_touch_delay(pi_e,Ce): #fonction calculant le touch delay. 
	if Ce==0: #si la célérité est nulle, 
		delta_h=math.inf #le délai pour toucher le bord/seuil est infini. Utilisation du module math pour être en mesure de manipuler l'infini. 
	elif Ce>0: 
		delta_h=(1-pi_e)/Ce
	elif Ce<0:
		delta_h=pi_e/(abs(Ce))
	return delta_h #delta_h le touch delay d'une variable. 

def slide_knocking_variable(graphe_etat,D,v,etat_courant,etat_suivant,dico_val_cele1,dico_val_cele2): #lafonction slide_knocking_variable vérifie si une variable ayant atteint son bord/seuil glisse ou non le long de ce bord/seuil. La fonction retourne True si la variable glisse, sinon elle retourne False. 
	Ce1=dico_val_cele1[v]
	sgn1=sgn(Ce1) #célérité de l'état courant.
	Ce2=dico_val_cele2[v]
	sgn2=sgn(Ce2)
	if sgn1*sgn2==-1: #si la variable ayant atteint son bord/seuil (knocking variable) rencontre un mur noir, 
		return True #renvoi de True car la variable v glisse sur son bord.
	elif (Ce1<0) and (etat_courant[v]==0): #si la variable ayant atteint son bord/seuil (knocking variable) est totalement dégradée, 
		return True #renvoi de True car la variable v glisse sur son bord.
	elif (Ce1>0) and (etat_courant[v]==D[v]): #si la variable ayant atteint son bord/seuil (knocking variable) est saturée,
		return True #renvoi de True car la variable v glisse sur son bord.
	else:
		return False #sinon renvoi de False car la variable ne glisse pas sur son bord. 

def make_dicovalcele(G,etat_init,M,D): #la fonction make_dicovalcele construit un dictionnaire. Elle associe à une variable : la valeur de sa célérité dans un état donné en argument de la fonction (etat_init).
	cele_init=celerite_par_etat(G,etat_init,M,D) #stockage dans cele_init, du dictionnaire associant à une variable sa celérité correspondante dans un état donné.
	dico_val_cele={}
	for v in cele_init: #pour chaque variable v dans le dictionaire cele_init, 
		for e in val_cele: #pour chaque célérité e dans le dictionnaire val_cele,
			if cele_init[v]==e: #si la célérité de v est égale à la célérité e,
				dico_val_cele[v]=val_cele[e] #alors à v est associée la valeur (flottant appartenant à l'ensemble des rééls positifs) de sa célérité dans un l'état fixé etat_init. 
	return dico_val_cele 

def first(etat_init,pi_init,var): #la fonction first retourne un tuple de la forme (first,dicot). first[0] stocke la variable qui atteint son bord en premier (sous forme de chaîne de caractère) dans un état donné (etat_init). first[1] stocke un dictionnaire associant à une variable, son touch_delay correspondant. 
	dico_val_cele=make_dicovalcele(G,etat_init,M,D) #dico_val_cele un dictionnaire associant à chaque variable v du graphe d'interaction, la valeur de sa célérité correspondante dans l'état etat_init entré en argument de la fonction,
	dicot={} #réinitialisation de dicot à chaque appel de la fonction first, 
	for v in var: #pour chaque variable v, 
		cele_v=dico_val_cele[v] #cele_v stocke la valeur de la célérité de v dans etat_init,
		delta_h=calcul_touch_delay(pi_init[v],cele_v) #calcul du touch_delay de v dans etat_init,
		if delta_h!=math.inf: #si le touch_delay de v est différent de plus l'infini, 
			dicot[v]=delta_h #on associe la valeur du touch delay calculé à la variable v,
		else:  
			dicot[v]=math.inf #(FLAG : à enlever et tester) sinon le touch_delay est égal à plus l'infini, 
	first_value=min(dicot.values()) #le touch_delay minimal correspond au touch_delay de la variable first stocké dans first_value. Notons que s'il existe plusieurs touch_delay minimaux égaux, le touch_delay minimal choisi est celui correspondant à la variable parcourue en premier dans la boucle for ligne 296.
	for e in dicot: #pour chaque variable e du graphe d'interaction,  
		if dicot[e]==first_value: #si le touch_delay de la variable e est égal au first_value, 
			first=e #alors la variable first (atteignant son bord en premier) est la variable e. 
	return first,dicot #first une chaîne de caractère et dicot un dictionnaire associant à une variable v, son touch delay. 

def etats_voisin(graphe_etat,etat_courant,first,dico_val_cele): #la fonction etats_voisin retourne l'état voisin sous forme de dictionnaire en sortie de l'état courant (dictionnaire en entrée) en fonction de la variable first (chaîne de caractère en entrée). La fonction prend également en entrée une liste de dictionnaire graphe_etat représentant l'ensemble des états pour un graphe d'interaction donné, ainsi qu'un dictionnaire dico_val_cele associant à une variable v, la valeur de sa célérité dans un état donné.
	ec=etat_courant.copy() #copie de l'état courant afin de conserver le contenu de l'état courant.
	del ec[first] #on suprimme la clé first, ainsi que sa valeur correspondante, du dictionnaire ec,
	voisin=[] #réinitialisation de voisin à chaque appel de la fonction, 
	for e in graphe_etat: #parcourt de chaque état du graphe d'état, 
		e1=e.copy() #copie de l'état e dans e1
		for s in e: #pour chaque sommet (une variable du graphe d'interaction) du dictionnaire e (un etat),
			if dico_val_cele[first]<0: #si la valeur de la célérité de la variable first est négative, 
				if (s==first) and (e[s]==etat_courant[s]-1): #si s est la variable first et si le niveau discret de s est décrémenté de 1, 
					del e1[s] #alors on supprime le niveau discret de la variable s (first) de l'état e1, 
					if (e1==ec): #si l'état e1 est égal à l'état ec (e1 et ec excluant le first),
						voisin=e #alors l'état voisin de la variable first est détecté sachant que la théorie de René Thomas se base sur la désynchronisation des transitions (passage d'un seuil à la fois) c'est pourquoi on ne considère qu'un état voisin à la fois,
			elif dico_val_cele[first]>0: #si la valeur de la célérité de la variable first est négative, 
				if (s==first) and (e[s]==etat_courant[s]+1): #si s est la variable first et si le niveau discret de s est incrémenté de 1, 
					del e1[s] #même principe que précédemment,
					if (e1==ec): 
						voisin=e
	return voisin

def new_position(var,var_copie,dico_val_cele,h0,first,dicot_etat): #fonction calculant la partie fractionnaire sortante à l'intérieur d'un état qualitatif à partir d'une partie fractionnaire entrante.
	etat_init=h0[0] #etat init l'état courant,
	pi_init=h0[1] #pi_init la coordonnée fractionnaire de chaque variable du graphe d'interaction,
	h=() 
	pi_next={} #initialisation de pi_next, associant à chaque variable, la coordonnée fractionnaire au pas de temps suivant (fixé par le touch_delay de la variable first),
	for v in var: #pour chaque variable du graphe d'interaction,
		cele_v=dico_val_cele[v] #cele_v le dictionnaire associant à chaque variable, sa valeur de célérité dans l'état courant,
		pi_n=(pi_init[v])+(dicot_etat[first]*cele_v) ##pi_n la coordonnée fractionnaire au pas suivant de la coordonnée fractionnaire. Avec pi_init[v] = la coordonnée fractionnaire en entrée de l'état courant, dicot_etat[first] = touch_delay de la variable first et cele_v = la valeur de la célérité de la variable v dans l'état courant,
		pi_next[v]=pi_n #stockage de la valeur de pi_n associé à la variable correspondante dans le dictionnaire pi_next,
		for vc in var_copie: #la liste des variales var est susceptible d'être modifiée dans le main. Dans ce cas, il faut s'assurer de manipuler la liste des variables intacte. vc est une copie de la liste intacte des variables. 
			if vc not in var: #si la liste var est modifiée par la suppression des variables glissantes (mur noir, saturation ou dégradation totale), alors récupération de ces variables dans la liste vc et, 
				pi_next[vc]=pi_init[vc] #la coordonnée fractionnaire de la variable glissante au pas suivant est égale à celle du pas précédent,
	h=(etat_init,pi_next) #en sortie, construction du tuple h avec h[0] l'état courant et pi_next la nouvelle coordonnée fractionnaire de chaque variable.
	return h

def new_position_new_state(dico_val_cele,etat_voisin,pi_next,first): #fonction calculant la partie fractionnaire entrante de l'état qualitatif voisin à partir de la partie fractionnaire sortante de l'état qualitatif courant.
	h=() 
	signe=sgn(dico_val_cele[first]) #signe, le signe de la célérité de la variable first tel que : si célérité<0 alors signe=-1, si célérité>0 alors signe=+1 et si célérité=0 alors signe=0.
	pi_new_first=(1-signe)/2 #changement de coordonnée fractionnaire de la variable first : elle vaut 0 si la célérité de first > 0 et 1 si la célérité de first < 0,
	del pi_next[first] #on supprime la variable first de l'ensemble des coordonnées fractionnaires des variables du graphe, 
	pi_next[first]=pi_new_first #afin de l'actualiser avec la nouvelle valeur calculée dans le nouvel état discret, 
	h=(etat_voisin,pi_next) #en sortie, construction du tuple h avec h[0] l'état courant et pi_next la nouvelle coordonnée fractionnaire de chaque variable.
	return h 

def printinfile(var,h,t): #fonction écrivant dans un fichier de sortie les valeurs etat_courant + pi_courant de chaque variable à un temps t (le touch delay de la variable first passant le seuil).
	etat=h[0] #etat l'état courant,
	pi=h[1] #pi l'ensemble des coordonnées fractionnaires des variables du graphe d'interaction,
	f.write('\n' + str(t) + '\t') #écriture du temps t dans le fichier de sortie (FLAG : peu élégant, à changer),
	for v in var: #pour chaque variable v du graphe, 
		val=etat[v]+pi[v] #calcule de la somme de son niveau discret avec sa partie fractionnaire et, 
		f.write(str(val) + '\t') #écriture dans le fichier de sortie.
	f.write('\n')

##############################################################################################
################################## ENTRÉES DE L'UTILISATEUR ##################################
##############################################################################################

##### GRAPHE D'INTERACTIONS #####
G={"mPC":["PC"], "mG":["G"]}
#Autre exemple : G={"mPC":["PC"], "mG":["G"], "mGF":["GF"], "mPCF":["PCF"],"mX":["X"], "mL":["L"]}.

##### FORMULES ASSOCIÉES AUX MULTIPLEXES #####
M={"mPC":">=(G,1)","mG":"not(>=(PC,1))"}
#Autre exemple : M={"mPC":">=(G,1)","mG":"or(>=(L,1),not(>=(PC,1)))", "mX":">=(L,1)", "mL":"not(>=(X,1))", "mPCF":">=(GF,1)", "mGF":"or(>=(PC,1),not(>=(PCF,1)))"}.

##### NIVEAU DISCET MAXIMAL #####
D={"G":1,"PC":1} 
#Autre exemple :  D={"G":1,"PC":1,"GF":1,"PCF":1, "L":1, "X":1}

##### ÉTAT INITIAL #####
etat_init={'G':0,'PC':0} 

##### COORDONNÉES FRACTIONNAIRES INITIALES #####
pi_init={'G':0.0,'PC':0.0} #coordonnées fractionnaires initiales pour l'ensemble des variables du graphe d'interaction.

####################################################################################
################################## INITIALISATION ##################################
####################################################################################

t=0.0 #temps initial

tmax=sys.argv[1] #1er argument entré par l'utilisateur à l'exécution du programme. 
tmax=int(tmax) #FLAG : inutile
tmax=float(tmax) 

####### INITIALISATION #######
from os import path as os_path 
repertoire = os_path.abspath(os_path.split(__file__)[0]) #détection du répertoire courant.

graphe_etat=extraction(G,D) #graphe_etat, l'ensemble des états pour le graphe d'interaction G.
val_cele=liste_cele(G,M,D) #initialisation des valeurs de chaque célérité.
var=liste_variable(G) #extraction de la liste des variables du graphe G.
var_copie=liste_variable(G) #copie préalable de la liste des variables du graphe d'interaction G. 


if not os.path.isfile('simulation.txt'): #si le fichier de sortie n'existe pas déjà dans Le répertoire courant, 
	f = open('simulation.txt','w') #ouverture du fichier de sortie simulation.txt. Si le fichier n'existe pas, alors création de celui ci, 
	f.write('t' + '\t') #écriture de l'entête de colonne "t" dans le fichier de sortie, 
	for v in var: #pour chaque variable, 
		f.write(v + '\t') #écriture de l'entête avec chaque nom de variable en tête de colonne, 
	f.write('\n' + str(t) + '\t') 
	for v in var:
		ini=pi_init[v]+etat_init[v]
		f.write(str(ini) + '\t')
else: #sinon message d'erreur, le programme s'interrompt dans son exécution. 
	print('Le fichier existe déjà dans le répertoire courant. Veuillez le supprimer ou le renommer.')

h_courant=(etat_init,pi_init) #h_courant, un tuple avec h[0] = état initial du système et h[1] = coordonnées fractionnaires initales du système. 
